import { defineStore } from "pinia";
import {
  BucketPermissionService,
  BucketService,
  Constraint,
  Permission,
} from "@/client/s3proxy";
import type {
  BucketOut,
  BucketIn,
  BucketPermissionOut,
  BucketPermissionIn,
  BucketPermissionParameters,
} from "@/client/s3proxy";
import { useAuthStore } from "@/stores/users";

export const useBucketStore = defineStore({
  id: "buckets",
  state: () =>
    ({
      bucketMapping: {},
      ownPermissions: {},
      bucketPermissionsMapping: {},
    }) as {
      bucketMapping: Record<string, BucketOut>;
      ownPermissions: Record<string, BucketPermissionOut>;
      bucketPermissionsMapping: Record<string, BucketPermissionOut[]>;
    },
  getters: {
    buckets(): BucketOut[] {
      const tempList = Object.values(this.bucketMapping);
      tempList.sort((bucketA, bucketB) => {
        if (bucketA.owner_constraint) {
          return -1;
        }
        if (bucketB.owner_constraint) {
          return 1;
        }
        return bucketA.name > bucketB.name ? 1 : -1;
      });
      return tempList;
    },
    ownBucketsAndFullPermissions(): string[] {
      const authStore = useAuthStore();
      const names = this.buckets
        .filter((bucket) => bucket.owner === authStore.currentUID)
        .map((bucket) => bucket.name)
        .concat(
          Object.values(this.ownPermissions)
            .filter((perm) => perm.permission === Permission.READWRITE)
            .map((perm) => perm.bucket_name),
        );
      names.sort();
      return names;
    },
    getBucketPermissions(): (bucketName: string) => BucketPermissionOut[] {
      return (bucketName) => this.bucketPermissionsMapping[bucketName] ?? [];
    },
    permissionFeatureAllowed(): (bucketName: string) => boolean {
      return (bucketName) => {
        // If a permission for the bucket exist, then false
        if (this.ownPermissions[bucketName] != undefined) {
          return false;
        }
        // If the bucket doesn't exist, then false
        if (this.bucketMapping[bucketName] == undefined) {
          return false;
        }
        // If there is a constraint on the bucket, then false otherwise true
        return this.bucketMapping[bucketName].owner_constraint == null;
      };
    },
    writableBuckets(): BucketOut[] {
      return this.buckets.filter((bucket) => {
        if (this.ownPermissions[bucket.name] != undefined) {
          return this.ownPermissions[bucket.name].permission !== "READ";
        }
        // If the user owns the bucket, check the bucket constraint
        return (
          bucket.owner_constraint == null ||
          bucket.owner_constraint == Constraint.WRITE
        );
      });
    },
    writableBucket(): (bucketName: string) => boolean {
      return (bucketName) => {
        // If this is a foreign bucket, check that the user has write permission
        if (this.ownPermissions[bucketName] != undefined) {
          return this.ownPermissions[bucketName].permission !== "READ";
        }
        // If the user owns the bucket, check the bucket constraint
        if (this.bucketMapping[bucketName] != undefined) {
          return (
            this.bucketMapping[bucketName].owner_constraint == null ||
            this.bucketMapping[bucketName].owner_constraint == Constraint.WRITE
          );
        }
        return false;
      };
    },
    readableBucket(): (bucketName: string) => boolean {
      return (bucketName) => {
        // If this is a foreign bucket, check that the user has read permission
        if (this.ownPermissions[bucketName] != undefined) {
          return this.ownPermissions[bucketName].permission !== "WRITE";
        }
        // If the user owns the bucket, check the bucket constraint
        if (this.bucketMapping[bucketName] != null) {
          return (
            this.bucketMapping[bucketName].owner_constraint == null ||
            this.bucketMapping[bucketName].owner_constraint == Constraint.READ
          );
        }
        return false;
      };
    },
  },
  actions: {
    fetchOwnPermissions(
      onFinally?: () => void,
    ): Promise<BucketPermissionOut[]> {
      const authStore = useAuthStore();
      if (authStore.authenticated && !authStore.foreignUser) {
        if (Object.keys(this.ownPermissions).length > 0) {
          onFinally?.();
        }
        return BucketPermissionService.bucketPermissionListPermissionsPerUser(
          authStore.currentUID,
        )
          .then((permissions) => {
            const new_permissions: Record<string, BucketPermissionOut> = {};
            for (const perm of permissions) {
              new_permissions[perm.bucket_name] = perm;
            }
            this.ownPermissions = new_permissions;
            return permissions;
          })
          .finally(onFinally);
      }
      return Promise.reject("Unauthorized");
    },
    deleteOwnPermission(bucketName: string) {
      delete this.ownPermissions[bucketName];
      delete this.bucketMapping[bucketName];
    },
    fetchBuckets(onFinally?: () => void): Promise<BucketOut[]> {
      if (this.buckets.length > 0) {
        onFinally?.();
      }
      const userStore = useAuthStore();
      return BucketService.bucketListBuckets(userStore.currentUID)
        .then((buckets) => {
          for (const bucket of buckets) {
            this.bucketMapping[bucket.name] = bucket;
          }
          const userRepository = useAuthStore();
          userRepository.fetchUsernames(
            buckets
              .map((bucket) => bucket.owner)
              .filter((owner) => owner != userRepository.currentUID),
          );
          return buckets;
        })
        .finally(onFinally);
    },
    fetchBucket(
      bucketName: string,
      onFinally?: () => void,
    ): Promise<BucketOut> {
      if (this.bucketMapping[bucketName] != undefined) {
        onFinally?.();
      }
      return BucketService.bucketGetBucket(bucketName)
        .then((bucket) => {
          this.bucketMapping[bucket.name] = bucket;
          return bucket;
        })
        .finally(onFinally);
    },
    deleteBucket(bucketName: string): Promise<void> {
      return BucketService.bucketDeleteBucket(bucketName, true).then(() => {
        delete this.bucketMapping[bucketName];
      });
    },
    createBucket(bucket: BucketIn): Promise<BucketOut> {
      return BucketService.bucketCreateBucket(bucket).then((createdBucket) => {
        this.bucketMapping[createdBucket.name] = createdBucket;
        return createdBucket;
      });
    },
    fetchBucketPermissions(
      bucketName: string,
      onFinally?: () => void,
    ): Promise<BucketPermissionOut[]> {
      if (this.bucketPermissionsMapping[bucketName] != undefined) {
        onFinally?.();
      }
      return BucketPermissionService.bucketPermissionListPermissionsPerBucket(
        bucketName,
      )
        .then((permissions) => {
          this.bucketPermissionsMapping[bucketName] = permissions;
          return permissions;
        })
        .finally(onFinally);
    },
    deleteBucketPermission(bucketName: string, uid: string): Promise<void> {
      return BucketPermissionService.bucketPermissionDeletePermissionForBucket(
        bucketName,
        uid,
      ).then(() => {
        const userRepository = useAuthStore();
        if (uid == userRepository.currentUID) {
          this.deleteOwnPermission(bucketName);
        }
        if (this.bucketPermissionsMapping[bucketName] == undefined) {
          this.fetchBucketPermissions(bucketName);
        } else {
          this.bucketPermissionsMapping[bucketName] =
            this.bucketPermissionsMapping[bucketName].filter(
              (permission) => permission.uid != uid,
            );
        }
      });
    },
    createBucketPermission(
      permissionIn: BucketPermissionIn,
    ): Promise<BucketPermissionOut> {
      return BucketPermissionService.bucketPermissionCreatePermission(
        permissionIn,
      ).then((permission) => {
        if (
          this.bucketPermissionsMapping[permission.bucket_name] == undefined
        ) {
          this.fetchBucketPermissions(permission.bucket_name);
        } else {
          this.bucketPermissionsMapping[permission.bucket_name].push(
            permission,
          );
        }
        return permission;
      });
    },
    updateBucketPermission(
      bucketName: string,
      uid: string,
      permissionParams: BucketPermissionParameters,
    ): Promise<BucketPermissionOut> {
      return BucketPermissionService.bucketPermissionUpdatePermission(
        bucketName,
        uid,
        permissionParams,
      ).then((permissionOut) => {
        if (this.bucketPermissionsMapping[bucketName] == undefined) {
          this.fetchBucketPermissions(bucketName);
        } else {
          const index = this.bucketPermissionsMapping[bucketName].findIndex(
            (permission) => (permission.uid = uid),
          );
          if (index > -1) {
            this.bucketPermissionsMapping[bucketName][index] = permissionOut;
          } else {
            this.bucketPermissionsMapping[bucketName].push(permissionOut);
          }
        }
        return permissionOut;
      });
    },
  },
});
