import { defineStore } from "pinia";
import type {
  DevWorkflowExecutionIn,
  WorkflowExecutionIn,
  WorkflowExecutionOut,
} from "@/client/workflow";
import {
  WorkflowExecutionService,
  WorkflowExecutionStatus,
} from "@/client/workflow";
import { useAuthStore } from "@/stores/users";
import dayjs from "dayjs";

export const useWorkflowExecutionStore = defineStore({
  id: "workflow-executions",
  state: () =>
    ({
      executionMapping: {},
    }) as {
      executionMapping: Record<string, WorkflowExecutionOut>;
    },
  getters: {
    executions(): WorkflowExecutionOut[] {
      return Object.values(this.executionMapping);
    },
  },
  actions: {
    fetchExecutions(onFinally?: () => void): Promise<WorkflowExecutionOut[]> {
      if (Object.keys(this.executionMapping).length > 0) {
        onFinally?.();
      }
      const userStore = useAuthStore();
      return WorkflowExecutionService.workflowExecutionListWorkflowExecutions(
        userStore.currentUID,
      )
        .then((executions) => {
          const newMapping: Record<string, WorkflowExecutionOut> = {};
          for (const execution of executions) {
            newMapping[execution.execution_id] = execution;
          }
          this.executionMapping = newMapping;
          return executions;
        })
        .finally(onFinally);
    },
    fetchExecution(
      executionId: string,
      onFinally?: () => void,
    ): Promise<WorkflowExecutionOut> {
      if (this.executionMapping[executionId] != undefined) {
        onFinally?.();
      }
      return WorkflowExecutionService.workflowExecutionGetWorkflowExecution(
        executionId,
      )
        .then((execution) => {
          this.executionMapping[execution.execution_id] = execution;
          return execution;
        })
        .finally(onFinally);
    },
    deleteExecution(executionId: string): Promise<void> {
      return WorkflowExecutionService.workflowExecutionDeleteWorkflowExecution(
        executionId,
      ).then(() => {
        delete this.executionMapping[executionId];
      });
    },
    cancelExecution(executionId: string): Promise<void> {
      return WorkflowExecutionService.workflowExecutionCancelWorkflowExecution(
        executionId,
      ).then(() => {
        if (this.executionMapping[executionId] == undefined) {
          this.fetchExecution(executionId);
        } else {
          this.executionMapping[executionId].status =
            WorkflowExecutionStatus.CANCELED;
          this.executionMapping[executionId].end_time = dayjs().unix();
        }
      });
    },
    startExecution(
      executionIn: WorkflowExecutionIn,
    ): Promise<WorkflowExecutionOut> {
      return WorkflowExecutionService.workflowExecutionStartWorkflow(
        executionIn,
      ).then((execution) => {
        this.executionMapping[execution.execution_id] = execution;
        return execution;
      });
    },
    startDevExecution(
      executionIn: DevWorkflowExecutionIn,
    ): Promise<WorkflowExecutionOut> {
      return WorkflowExecutionService.workflowExecutionStartArbitraryWorkflow(
        executionIn,
      ).then((execution) => {
        this.executionMapping[execution.execution_id] = execution;
        return execution;
      });
    },
  },
});
